//
// Created by gwall on 1/9/18.
//


#include "encoder_helpers.h"

using namespace std;

int test_error(bool cond, const char *str, context_t &ctx) {
    if (cond) {
        cerr << str << endl;
        return cleanup(ctx, 1);
    }
    return 0;
}

int cleanup(context_t &ctx, int error) {
    if (ctx.enc && ctx.enc->isInError())
    {
        cerr << "Encoder is in error" << endl;
        error = 1;
    }
    if (ctx.got_error)
    {
        error = 1;
    }

    if (ctx.pBitStreamCrc)
    {
        char *pgold_crc = ctx.gold_crc;
        Crc *pout_crc= ctx.pBitStreamCrc;
        char StrCrcValue[20];
        snprintf (StrCrcValue, 20, "%u", pout_crc->CrcValue);
        // Remove CRLF from end of CRC, if present
        do {
            unsigned int len = strlen(pgold_crc);
            if (len == 0) break;
            if (pgold_crc[len-1] == '\n')
                pgold_crc[len-1] = '\0';
            else if (pgold_crc[len-1] == '\r')
                pgold_crc[len-1] = '\0';
            else
                break;
        } while(1);

        if (strcmp (StrCrcValue, pgold_crc))
        {
            cout << "======================" << endl;
            cout << "video_encode: CRC FAILED" << endl;
            cout << "======================" << endl;
            cout << "Encoded CRC: " << StrCrcValue << " Gold CRC: " << pgold_crc << endl;
            error = 1;
        }
        else
        {
            cout << "======================" << endl;
            cout << "video_encode: CRC PASSED" << endl;
            cout << "======================" << endl;
        }

        CloseCrc(&ctx.pBitStreamCrc);
    }

    delete ctx.enc;
    delete ctx.in_file;
    delete ctx.out_file;
    delete ctx.roi_Param_file;
    delete ctx.recon_Ref_file;
    delete ctx.rps_Param_file;
    delete ctx.hints_Param_file;
    delete ctx.gdr_Param_file;
    delete ctx.gdr_out_file;

    if (ctx.in_file_path)
        free(ctx.in_file_path);

    if (ctx.out_file_path)
        free(ctx.out_file_path);

    free(ctx.ROI_Param_file_path);
    free(ctx.Recon_Ref_file_path);
    free(ctx.RPS_Param_file_path);
    free(ctx.hints_Param_file_path);
    free(ctx.GDR_Param_file_path);
    free(ctx.GDR_out_file_path);
    delete ctx.runtime_params_str;

    if (error)
    {
        cout << "App run failed" << endl;
    }
    else
    {
        cout << "App run was successful" << endl;
    }
    return -error;
}


void abort(context_t *ctx)
{
    ctx->got_error = true;
    ctx->enc->abort();
}

//Initialise CRC Rec and creates CRC Table based on the polynomial.
Crc* InitCrc(unsigned int CrcPolynomial)
{
    unsigned short int i;
    unsigned short int j;
    unsigned int tempcrc;
    Crc *phCrc;
    phCrc = (Crc*) malloc (sizeof(Crc));
    if (phCrc == nullptr)
    {
        cerr << "Mem allocation failed for Init CRC" <<endl;
        return nullptr;
    }

    memset (phCrc, 0, sizeof(Crc));

    for (i = 0; i <= 255; i++)
    {
        tempcrc = i;
        for (j = 8; j > 0; j--)
        {
            if (tempcrc & 1)
            {
                tempcrc = (tempcrc >> 1) ^ CrcPolynomial;
            }
            else
            {
                tempcrc >>= 1;
            }
        }
        phCrc->CRCTable[i] = tempcrc;
    }

    phCrc->CrcValue = 0;
    return phCrc;
}

//Calculates CRC of data provided in by buffer.
void CalculateCrc(Crc *phCrc, unsigned char *buffer, uint32_t count)
{
    unsigned char *p;
    unsigned int temp1;
    unsigned int temp2;
    unsigned int crc = phCrc->CrcValue;
    unsigned int *CRCTable = phCrc->CRCTable;

    if(!count)
        return;

    p = (unsigned char *) buffer;
    while (count-- != 0)
    {
        temp1 = (crc >> 8) & 0x00FFFFFFL;
        temp2 = CRCTable[((unsigned int) crc ^ *p++) & 0xFF];
        crc = temp1 ^ temp2;
    }

    phCrc->CrcValue = crc;
}

//Closes CRC related handles.
void CloseCrc(Crc **phCrc)
{
    if (*phCrc)
        free (*phCrc);
}

int write_encoder_output_frame(ofstream * stream, NvBuffer * buffer)
{
    stream->write((char *) buffer->planes[0].data, buffer->planes[0].bytesused);
    return 0;
}

bool encoder_capture_plane_dq_callback(struct v4l2_buffer *v4l2_buf, NvBuffer * buffer,
                                  NvBuffer * shared_buffer, void *arg)
{
    context_t *ctx = (context_t *) arg;
    NvVideoEncoder *enc = ctx->enc;
    uint32_t frame_num = ctx->enc->capture_plane.getTotalDequeuedBuffers() - 1;
    uint32_t ReconRef_Y_CRC = 0;
    uint32_t ReconRef_U_CRC = 0;
    uint32_t ReconRef_V_CRC = 0;
    static uint32_t num_encoded_frames = 1;

    if (v4l2_buf == nullptr)
    {
        cout << "Error while dequeing buffer from output plane" << endl;
        abort(ctx);
        return false;
    }

    // GOT EOS from encoder. Stop dqthread.
    if (buffer->planes[0].bytesused == 0)
    {
        return false;
    }

    // Computing CRC with each frame
    if(ctx->pBitStreamCrc)
        CalculateCrc (ctx->pBitStreamCrc, buffer->planes[0].data, buffer->planes[0].bytesused);

    write_encoder_output_frame(ctx->out_file, buffer);
    num_encoded_frames++;

    // Accounting for the first frame as it is only sps+pps
    if (ctx->gdr_out_frame_number != 0xFFFFFFFF)
        if ( (ctx->enableGDR) && (ctx->GDR_out_file_path) && (num_encoded_frames >= ctx->gdr_out_frame_number+1))
            write_encoder_output_frame(ctx->gdr_out_file, buffer);

    if (ctx->report_metadata)
    {
        v4l2_ctrl_videoenc_outputbuf_metadata enc_metadata;
        if (ctx->enc->getMetadata(v4l2_buf->index, enc_metadata) == 0)
        {
            if (ctx->bReconCrc && enc_metadata.bValidReconCRC) {
                // CRC for Recon frame
                cout << "Frame: " << frame_num << endl;
                cout << "ReconFrame_Y_CRC " << enc_metadata.ReconFrame_Y_CRC <<
                     " ReconFrame_U_CRC " << enc_metadata.ReconFrame_U_CRC <<
                     " ReconFrame_V_CRC " << enc_metadata.ReconFrame_V_CRC <<
                     endl;

                if (!ctx->recon_Ref_file->eof()) {
                    *ctx->recon_Ref_file >> ReconRef_Y_CRC;
                    *ctx->recon_Ref_file >> ReconRef_U_CRC;
                    *ctx->recon_Ref_file >> ReconRef_V_CRC;
                }

                if ((ReconRef_Y_CRC != enc_metadata.ReconFrame_Y_CRC) ||
                    (ReconRef_U_CRC != enc_metadata.ReconFrame_U_CRC) ||
                    (ReconRef_V_CRC != enc_metadata.ReconFrame_V_CRC))
                {
                    cout << "Recon CRC FAIL" << endl;
                    cout << "ReconRef_Y_CRC " << ReconRef_Y_CRC <<
                         " ReconRef_U_CRC " << ReconRef_U_CRC <<
                         " ReconRef_V_CRC " << ReconRef_V_CRC <<
                         endl;
                    abort(ctx);
                    return false;
                }
                cout << "Recon CRC PASS for frame : " << frame_num << endl;
            } else if (ctx->externalRPS && enc_metadata.bRPSFeedback_status) {
                // RPS Feedback
                cout << "Frame: " << frame_num << endl;
                cout << "nCurrentRefFrameId " << enc_metadata.nCurrentRefFrameId <<
                     " nActiveRefFrames " << enc_metadata.nActiveRefFrames << endl;

                for (uint32_t i = 0; i < enc_metadata.nActiveRefFrames; i++)
                {
                    cout << "FrameId " << enc_metadata.RPSList[i].nFrameId <<
                         " IdrFrame " <<  (int) enc_metadata.RPSList[i].bIdrFrame <<
                         " LTRefFrame " <<  (int) enc_metadata.RPSList[i].bLTRefFrame <<
                         " PictureOrderCnt " << enc_metadata.RPSList[i].nPictureOrderCnt <<
                         " FrameNum " << enc_metadata.RPSList[i].nFrameNum <<
                         " LTFrameIdx " <<  enc_metadata.RPSList[i].nLTRFrameIdx << endl;
                }
            } else if (ctx->externalRCHints) {
                // Rate Control Feedback
                cout << "Frame: " << frame_num << endl;
                cout << "EncodedBits " << enc_metadata.EncodedFrameBits <<
                     " MinQP " << enc_metadata.FrameMinQP <<
                     " MaxQP " << enc_metadata.FrameMaxQP <<
                     endl;
            } else {
                cout << "Frame " << frame_num <<
                     ": isKeyFrame=" << (int) enc_metadata.KeyFrame <<
                     " AvgQP=" << enc_metadata.AvgQP <<
                     " MinQP=" << enc_metadata.FrameMinQP <<
                     " MaxQP=" << enc_metadata.FrameMaxQP <<
                     " EncodedBits=" << enc_metadata.EncodedFrameBits <<
                     endl;
            }
        }
    }
    if (ctx->dump_mv)
    {
        v4l2_ctrl_videoenc_outputbuf_metadata_MV enc_mv_metadata;
        if (ctx->enc->getMotionVectors(v4l2_buf->index, enc_mv_metadata) == 0)
        {
            uint32_t numMVs = enc_mv_metadata.bufSize / sizeof(MVInfo);
            MVInfo *pInfo = enc_mv_metadata.pMVInfo;

            cout << "Frame " << frame_num << ": Num MVs=" << numMVs << endl;

            for (uint32_t i = 0; i < numMVs; i++, pInfo++)
            {
                cout << i << ": mv_x=" << pInfo->mv_x <<
                     " mv_y=" << pInfo->mv_y <<
                     " weight=" << pInfo->weight <<
                     endl;
            }
        }
    }

    if (enc->capture_plane.qBuffer(*v4l2_buf, nullptr) < 0)
    {
        cerr << "Error while Qing buffer at capture plane" << endl;
        abort(ctx);
        return false;
    }

    return true;
}

int setup_encoder(context_t& ctx) {
    int ret = 0;

/*
    if (ctx.runtime_params_str)
    {
        get_next_runtime_param_change_frame(&ctx);
    }
*/
    if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H265)
    {
        test_error(ctx.width < 144 || ctx.height < 144, "Height/Width should be"
                " > 144 for H.265", ctx);
    }

    if (ctx.use_gold_crc)
    {
        ctx.pBitStreamCrc = InitCrc(CRC32_POLYNOMIAL);
        test_error(!ctx.pBitStreamCrc, "InitCrc failed", ctx);
    }

    ctx.in_file = new ifstream(ctx.in_file_path);
    test_error(!ctx.in_file->is_open(), "Could not open input file", ctx);

    ctx.out_file = new ofstream(ctx.out_file_path);
    test_error(!ctx.out_file->is_open(), "Could not open output file", ctx);

    if (ctx.ROI_Param_file_path) {
        ctx.roi_Param_file = new ifstream(ctx.ROI_Param_file_path);
        test_error(!ctx.roi_Param_file->is_open(), "Could not open roi param file", ctx);
    }

    if (ctx.Recon_Ref_file_path) {
        ctx.recon_Ref_file = new ifstream(ctx.Recon_Ref_file_path);
        test_error(!ctx.recon_Ref_file->is_open(), "Could not open recon crc reference file", ctx);
    }

    if (ctx.RPS_Param_file_path) {
        ctx.rps_Param_file = new ifstream(ctx.RPS_Param_file_path);
        test_error(!ctx.rps_Param_file->is_open(), "Could not open rps param file", ctx);
    }

    if (ctx.GDR_Param_file_path) {
        ctx.gdr_Param_file = new ifstream(ctx.GDR_Param_file_path);
        test_error(!ctx.gdr_Param_file->is_open(), "Could not open GDR param file", ctx);
    }

    if (ctx.GDR_out_file_path) {
        ctx.gdr_out_file = new ofstream(ctx.GDR_out_file_path);
        test_error(!ctx.gdr_out_file->is_open(), "Could not open GDR Out file", ctx);
    }

    if (ctx.hints_Param_file_path) {
        ctx.hints_Param_file = new ifstream(ctx.hints_Param_file_path);
        test_error(!ctx.hints_Param_file->is_open(), "Could not open hints param file", ctx);
    }

    ctx.enc = NvVideoEncoder::createVideoEncoder("enc0");
    test_error(!ctx.enc, "Could not create encoder", ctx);

    // It is necessary that Capture Plane format be set before Output Plane
    // format.
    // Set encoder capture plane format. It is necessary to set width and
    // height on thr capture plane as well
    /*
    ret = ctx.enc->setCapturePlaneFormat(ctx.encoder_pixfmt, ctx.width,
                                         ctx.height, 2 * 1024 * 1024);
                                         */

    // https://devtalk.nvidia.com/default/topic/1027631/formatting-images-to-feed-into-nvvideoencoder-tegra-multimedia-api-/
    ret = ctx.enc->setCapturePlaneFormat(ctx.encoder_pixfmt, ctx.width,
                                         ctx.height, (ctx.width * ctx.height * 3) / 2);
    test_error(ret < 0, "Could not set output plane format", ctx);

    // Set encoder output plane format
    ret = ctx.enc->setOutputPlaneFormat(V4L2_PIX_FMT_YUV420M, ctx.width, ctx.height);
    test_error(ret < 0, "Could not set output plane format", ctx);

    ret = ctx.enc->setBitrate(ctx.bitrate);
    test_error(ret < 0, "Could not set encoder bitrate", ctx);

    ret = ctx.enc->setProfile(ctx.profile);
    test_error(ret < 0, "Could not set encoder profile", ctx);

    if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H264)
    {
        ret = ctx.enc->setLevel(ctx.level);
        test_error(ret < 0, "Could not set encoder level", ctx);
    }

    ret = ctx.enc->setRateControlMode(ctx.ratecontrol);
    test_error(ret < 0, "Could not set encoder rate control mode", ctx);

    ret = ctx.enc->setIDRInterval(ctx.idr_interval);
    test_error(ret < 0, "Could not set encoder IDR interval", ctx);

    ret = ctx.enc->setIFrameInterval(ctx.iframe_interval);
    test_error(ret < 0, "Could not set encoder I-Frame interval", ctx);

    ret = ctx.enc->setFrameRate(ctx.fps_n, ctx.fps_d);
    test_error(ret < 0, "Could not set framerate", ctx);

    if (ctx.temporal_tradeoff_level)
    {
        ret = ctx.enc->setTemporalTradeoff(ctx.temporal_tradeoff_level);
        test_error(ret < 0, "Could not set temporal tradeoff level", ctx);
    }

    if (ctx.slice_length)
    {
        ret = ctx.enc->setSliceLength(ctx.slice_length_type,
                                      ctx.slice_length);
        test_error(ret < 0, "Could not set slice length params", ctx);
    }

    if (ctx.hw_preset_type)
    {
        ret = ctx.enc->setHWPresetType(ctx.hw_preset_type);
        test_error(ret < 0, "Could not set encoder HW Preset Type", ctx);
    }

    if (ctx.virtual_buffer_size)
    {
        ret = ctx.enc->setVirtualBufferSize(ctx.virtual_buffer_size);
        test_error(ret < 0, "Could not set virtual buffer size", ctx);
    }

    if (ctx.num_reference_frames)
    {
        ret = ctx.enc->setNumReferenceFrames(ctx.num_reference_frames);
        test_error(ret < 0, "Could not set num reference frames", ctx);
    }

    if (ctx.slice_intrarefresh_interval)
    {
        ret = ctx.enc->setSliceIntrarefresh(ctx.slice_intrarefresh_interval);
        test_error(ret < 0, "Could not set slice intrarefresh interval", ctx);
    }

    if (ctx.insert_sps_pps_at_idr)
    {
        ret = ctx.enc->setInsertSpsPpsAtIdrEnabled(true);
        test_error(ret < 0, "Could not set insertSPSPPSAtIDR", ctx);
    }

    if (ctx.num_b_frames != (uint32_t) -1)
    {
        ret = ctx.enc->setNumBFrames(ctx.num_b_frames);
        test_error(ret < 0, "Could not set number of B Frames", ctx);
    }

    if ((ctx.nMinQpI != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMaxQpI != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMinQpP != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMaxQpP != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMinQpB != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMaxQpB != (uint32_t)QP_RETAIN_VAL))
    {
        ret = ctx.enc->setQpRange(ctx.nMinQpI, ctx.nMaxQpI, ctx.nMinQpP,
                                  ctx.nMaxQpP, ctx.nMinQpB, ctx.nMaxQpB);
        test_error(ret < 0, "Could not set quantization parameters", ctx);
    }

    if (ctx.dump_mv)
    {
        ret = ctx.enc->enableMotionVectorReporting();
        test_error(ret < 0, "Could not enable motion vector reporting", ctx);
    }

    if (ctx.bnoIframe) {
        ctx.iframe_interval = ((1<<31) + 1); /* TODO: how can we do this properly */
        ret = ctx.enc->setIFrameInterval(ctx.iframe_interval);
        test_error(ret < 0, "Could not set encoder I-Frame interval", ctx);
    }

    if (ctx.enableROI) {
        v4l2_enc_enable_roi_param VEnc_enable_ext_roi_ctrl;

        VEnc_enable_ext_roi_ctrl.bEnableROI = ctx.enableROI;
        ret = ctx.enc->enableROI(VEnc_enable_ext_roi_ctrl);
        test_error(ret < 0, "Could not enable ROI", ctx);
    }

    if (ctx.bReconCrc) {
        v4l2_enc_enable_reconcrc_param VEnc_enable_recon_crc_ctrl;

        VEnc_enable_recon_crc_ctrl.bEnableReconCRC = ctx.bReconCrc;
        ret = ctx.enc->enableReconCRC(VEnc_enable_recon_crc_ctrl);
        test_error(ret < 0, "Could not enable Recon CRC", ctx);
    }

    if (ctx.externalRPS) {
        v4l2_enc_enable_ext_rps_ctr VEnc_enable_ext_rps_ctrl;

        VEnc_enable_ext_rps_ctrl.bEnableExternalRPS = ctx.externalRPS;
        if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H264) {
            VEnc_enable_ext_rps_ctrl.bGapsInFrameNumAllowed = ctx.bGapsInFrameNumAllowed;
            VEnc_enable_ext_rps_ctrl.nH264FrameNumBits = ctx.nH264FrameNumBits;
        }
        if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H265) {
            VEnc_enable_ext_rps_ctrl.nH265PocLsbBits = ctx.nH265PocLsbBits;
        }
        ret = ctx.enc->enableExternalRPS(VEnc_enable_ext_rps_ctrl);
        test_error(ret < 0, "Could not enable external RPS", ctx);
    }

    if (ctx.externalRCHints) {
        v4l2_enc_enable_ext_rate_ctr VEnc_enable_ext_rate_ctrl;

        VEnc_enable_ext_rate_ctrl.bEnableExternalPictureRC = ctx.externalRCHints;
        VEnc_enable_ext_rate_ctrl.nsessionMaxQP = ctx.sMaxQp;

        ret = ctx.enc->enableExternalRC(VEnc_enable_ext_rate_ctrl);
        test_error(ret < 0, "Could not enable external RC", ctx);
    }

    // Query, Export and Map the output plane buffers so that we can read
    // raw data into the buffers
    ret = ctx.enc->output_plane.setupPlane(V4L2_MEMORY_MMAP, 10, true, false);
    test_error(ret < 0, "Could not setup output plane", ctx);

    // Query, Export and Map the output plane buffers so that we can write
    // encoded data from the buffers
    ret = ctx.enc->capture_plane.setupPlane(V4L2_MEMORY_MMAP, 10, true, false);
    test_error(ret < 0, "Could not setup capture plane", ctx);

    // output plane STREAMON
    ret = ctx.enc->output_plane.setStreamStatus(true);
    test_error(ret < 0, "Error in output plane streamon", ctx);

    // capture plane STREAMON
    ret = ctx.enc->capture_plane.setStreamStatus(true);
    test_error(ret < 0, "Error in capture plane streamon", ctx);

    ctx.enc->capture_plane.
            setDQThreadCallback(encoder_capture_plane_dq_callback);

    // startDQThread starts a thread internally which calls the
    // encoder_capture_plane_dq_callback whenever a buffer is dequeued
    // on the plane
    ctx.enc->capture_plane.startDQThread(&ctx);

    // Enqueue all the empty capture plane buffers
    for (uint32_t i = 0; i < ctx.enc->capture_plane.getNumBuffers(); i++)
    {
        struct v4l2_buffer v4l2_buf;
        struct v4l2_plane planes[MAX_PLANES];

        memset(&v4l2_buf, 0, sizeof(v4l2_buf));
        memset(planes, 0, MAX_PLANES * sizeof(struct v4l2_plane));

        v4l2_buf.index = i;
        v4l2_buf.m.planes = planes;

        ret = ctx.enc->capture_plane.qBuffer(v4l2_buf, nullptr);
        if (ret < 0)
        {
            cerr << "Error while queueing buffer at capture plane" << endl;
            abort(&ctx);
            //goto cleanup;
            return cleanup(ctx, 1);
        }
    }

    return ret;
}
